/*
  cotest - a program to determine why you can't log into a slice

  authors: Vivek Pai, KyoungSoo Park

*/

#define VERSION "0.8a"

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <netdb.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <time.h>
#include <unistd.h>

#ifdef PSEPR_ALIVE
static char *allAlerts;
#endif

static char *servName;
static char *motd;
static char cacheDir[256];

static int debug = 0;

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

#ifndef SUCCESS
#define SUCCESS 0
#endif

#ifndef FAILURE
#define FAILURE (-1)
#endif

/*-----------------------------------------------------------------*/
static char *
StrdupLower(const char *orig)
{
  char *temp;
  int i;

  if ((temp = strdup(orig)) == NULL) {
    fprintf(stderr, "no memory in strduplower\n");
    exit(-1);
  }
  for (i = 0; temp[i]; i++) {
    if (isupper((int) temp[i]))
      temp[i] = tolower(temp[i]);
  }
  return(temp);
}
/*-----------------------------------------------------------------*/
static char *
GetNextLineBack(FILE *file, int lower)
{
  /* reads the next non-blank line of the file. strips off any leading
     space, any comments, and any trailing space.  returns a lowercase
     version of the line that has been malloc'd */
  char line[1024];

  while (fgets(line, sizeof(line), file) != NULL) {
    char *temp;
    int len;

    /* strip off any comments, leading and trailing whitespace */
    if ((temp = strchr(line, '#')) != NULL)
      *temp = 0;
    len = strlen(line);
    while (len > 0 && isspace((int) line[len-1])) {
      len--;
      line[len] = 0;
    }
    temp = line;
    while (isspace((int) *temp))
      temp++;
    if (temp[0] == 0)
      continue;			/* blank line, move on */

    if (lower)
      return(StrdupLower(temp));
    return(strdup(temp));
  }

  return(NULL);
}
/*-----------------------------------------------------------------*/
static char *
GetLowerNextLine(FILE *file)
{
  return(GetNextLineBack(file, TRUE));
}
/*-----------------------------------------------------------------*/
static char *
GetURL(char *host, int portNum, char *uri, int waitSecs, int *retCode)
{
  int fd;
  int dummy;
  struct hostent *ent;
  struct in_addr resultAddr;
  struct sockaddr_in saddr;
  struct pollfd pfd;
  char req[8192];
  int bufAlloc = 0;
  int bufUsed = 0;
  char *buf = NULL;
  int res;

  if (retCode == NULL)
    retCode = &dummy;
  *retCode = -1;

  if ((ent = gethostbyname(host)) == NULL)
    return(NULL);
  resultAddr = *((struct in_addr *)(ent->h_addr));

  /* start connection */
  memset(&saddr, 0, sizeof(saddr));
  saddr.sin_family = AF_INET;
  saddr.sin_addr = resultAddr;
  saddr.sin_port = htons(portNum);

  /* get socket */
  if ((fd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    fprintf(stderr, "couldn't create socket\n");
    exit(-1);
  }

  /* make it async */
  if (fcntl(fd, F_SETFL, O_NDELAY) == -1) {
    fprintf(stderr, "couldn't make sock async\n");
    exit(-1);
  }

  /* start connection */
  if (connect(fd, (struct sockaddr *) &saddr, 
	      sizeof(struct sockaddr_in)) < 0) {
    if (errno != EINPROGRESS) {
      close(fd);
      return(NULL);
    }
  }

  /* wait for answer */
  pfd.fd = fd;
  pfd.events = POLLOUT;
  if ((res = poll(&pfd, 1, (waitSecs > 0) ? waitSecs * 1000 : -1)) < 1) {
    close(fd);
    return(NULL);
  }

  sprintf(req, 
	  "GET /%s HTTP/1.0\r\n"
	  "Host: %s:%d\r\n"
	  "Connection: close\r\n"
	  "\r\n",
	  uri, host, portNum);

  if (write(fd, req, strlen(req)) < 0) {
    close(fd);
    return(NULL);
  }

  pfd.events = POLLIN;

  while (1) {
    if ((res = poll(&pfd, 1, (waitSecs > 0) ? waitSecs * 1000 : -1)) < 1) {
      close(fd);
      *retCode = bufUsed;
      return(buf);
    }
    
    if (bufUsed >= bufAlloc-1) {
      bufAlloc *= 2;
      if (bufAlloc < 2048)
	bufAlloc = 2048;
      if ((buf = realloc(buf, bufAlloc)) == NULL) {
	fprintf(stderr, "failed realloc buf\n");
	exit(-1);
      }
    }

    if ((res = read(fd, &buf[bufUsed], bufAlloc-bufUsed-1)) <= 0) {
      close(fd);
      *retCode = bufUsed;
      return(buf);
    }
    bufUsed += res;
    buf[bufUsed] = 0;
  }

}
/*-----------------------------------------------------------------*/
#ifdef PSEPR_ALIVE
static char *
Match(char *big, char *little, int destroy)
{
  char *walk = big;
  char *dest, *destStart;

  if (big == NULL)
    return(NULL);
  if (little == NULL || little[0] == '\0') {
    fprintf(stderr, "Match called with invalid little\n");
    exit(-1);
  }

  if ((dest = destStart = malloc(1+strlen(big))) == NULL) {
    fprintf(stderr, "failed in allocating match buffer\n");
    exit(-1);
  }
  *dest = '\0';

  while (1) {
    char *end;
    while (*walk == '\n')
      walk++;
    if (*walk == '\0')
      break;
    for (end = walk; *end != '\0' && *end != '\n'; end++)
      ;

    /* optimistically copy the current line */
    memcpy(dest, walk, end-walk);
    dest[end-walk] = 0;

    /* if matches, keep the current line, otherwise remove */
    if (strstr(dest, little) != NULL) {
      dest += end-walk;
      *dest++ = '\n';
      *dest++ = '\0';
    }
    else
      *dest = 0;
    walk = end;
  }
  if (destroy)
    free(big);
  dest = (destStart[0] == '\0') ? NULL : strdup(destStart);
  free(destStart);
  return(dest);
}
#endif
/*-----------------------------------------------------------------*/
static char *
FetchCacheableFile(char *path, char *file, 
		   int maxAge, char *hostName, char *uri)
{
  char *buf = NULL;
  char filename[1024];
  int skipFetch = FALSE;

  sprintf(filename, "%s/%s", path, file);

  if (access(filename, R_OK) == 0) {
    struct stat sb;
    if (stat(filename, &sb) != 0 || sb.st_size < 1) 
      fprintf(stderr, "couldn't stat %s - weird, but will refetch\n", 
	      file);
    else if (abs(sb.st_mtime - time(NULL)) > maxAge) {
      if (debug)
	fprintf(stderr, "%s stale - needs fetching\n", file);
    }
    else {
      FILE *f = fopen(filename, "r");
      if (f == NULL)
	fprintf(stderr, "couldn't open %s\n", file);
      else {
	buf = malloc(1+sb.st_size);
	buf[sb.st_size] = '\0';
	if (fread(buf, 1, sb.st_size, f) != sb.st_size)
	  fprintf(stderr, "%s incompletely read\n", file);
	else {
	  if (debug)
	    fprintf(stderr, "cached %s read successfully\n", file);
	  skipFetch = TRUE;
	}
	fclose(f);
      }
    }
  }
  else {
    if (debug)
      fprintf(stderr, "no %s found - needs fetching\n", file);
  }

  if (!skipFetch) {
    FILE *f;
    if (debug)
      fprintf(stderr, "fetching %s\n", file);
    free(buf);
    if ((buf = GetURL(hostName, 80, uri, 60, NULL)) == NULL) {
      fprintf(stderr, "unable to get http://%s/%s\n", hostName, uri);
      exit(-1);
    }
    if ((f = fopen(filename, "w")) == NULL)
      fprintf(stderr, "couldn't write %s\n", file);
    else {
      fwrite(buf, 1, strlen(buf), f);
      fclose(f);
    }
  }

  return(buf);
}
/*-----------------------------------------------------------------*/
static int
CoMonTests(char *nodeName)
{
  int cotopAge = -1;
  double respTime = 0;
  double load1 = 0;
  double drift = -1;
  int lastSsh = -999999;
  int result = FAILURE;
  char uri[256];
  char *comonRes;
  char *temp;

  /* first see what CoMon thinks of it */
  sprintf(uri, "status/mon_byname/%s", nodeName);
  if ((comonRes = 
       FetchCacheableFile(cacheDir, nodeName, 
			  300, servName, uri)) == NULL) {
    fprintf(stderr, "%s: no results in CoMon\n", nodeName);
    return(FAILURE);
  }
  
  if ((temp = strstr(comonRes, "CotopAge: ")) != NULL)
    cotopAge = atoi(temp+strlen("CotopAge: "));
  if ((temp = strstr(comonRes, "RespTime: ")) != NULL)
    respTime = atof(temp+strlen("RespTime: "));
  if ((temp = strstr(comonRes, "Load1: ")) != NULL)
    load1 = atof(temp+strlen("Load1: "));
  if ((temp = strstr(comonRes, "Drift: ")) != NULL)
    drift = fabs(atof(temp+strlen("Drift: ")));
  if ((temp = strstr(comonRes, "LastSsh: ")) != NULL)
    lastSsh = time(NULL) - atoi(temp+strlen("LastSsh: "));
  if (load1 > 100 && drift > 1000) {
    fprintf(stderr, 
	    "%s: heavily loaded, probably frozen - "
	    "load %.0f, drift %.0f sec\n", 
	    nodeName, load1, drift);
  }
  else if (respTime == 0) {
    if (cotopAge == -1 || cotopAge > 1000)
      fprintf(stderr, "%s: both CoMon and CoTop think node is dead\n", 
	      nodeName);
    else {
      fprintf(stderr, "%s: CoMon failing, but CoTop ok\n", nodeName);
      result = SUCCESS;
    }
  }
  else {
    if (lastSsh != -999999 && abs(lastSsh) > 3600) {
      if (lastSsh > 0) {
	fprintf(stderr, 
		"%s: last ssh %.1f hours ago, very suspicious\n", 
		nodeName, lastSsh/3600.0);
      }
      else {
	fprintf(stderr, 
		"%s: your clock may be off - last ssh %.1f hours in the future\n", 
		nodeName, -lastSsh/3600.0);
      }
    }
    else if (cotopAge == -1 || cotopAge > 1000) {
      if (drift > 1000)
	fprintf(stderr, 
		"%s: probably frozen - CoMon running, "
		"CoTop failing, drift %.0f secs\n", nodeName, drift);
      else {
	fprintf(stderr, 
		"%s: slightly suspicious - CoMon running, CoTop failing\n", 
		nodeName);
	result = SUCCESS;
      }
    }
    else {
      result = SUCCESS;
      if (drift > 1000)
	fprintf(stderr, 
		"%s: just fyi - clock drift %.0f secs\n", nodeName, drift);
      else if (load1 > 100)
	fprintf(stderr, "%s: heavily loaded - load %.0f\n", nodeName, load1);
      else {
	if (lastSsh != -999999)
	  fprintf(stderr, "%s: CoMon thinks the node's OK and ssh-able\n", 
		  nodeName);
	else
	  fprintf(stderr, "%s: CoMon thinks the node's OK\n", nodeName);
      }
    }
  }
  free(comonRes);
  return(result);
}
/*-----------------------------------------------------------------*/
#ifdef PSEPR_ALIVE
static int
AlertTests(char *nodeName,char *sliceName)
{
  char *bad = NULL;
  int result = FAILURE;
  char *alerts;

  /* check if node has any outstanding alerts */
  if ((alerts = Match(allAlerts, nodeName, FALSE)) == NULL)
    return(SUCCESS);

  if ((bad = Match(alerts, "node down", FALSE)) != NULL)
    fprintf(stderr, "%s: PEPR thinks node down\n", nodeName);
  else if ((bad = Match(alerts, "failed ssh.22", FALSE)) != NULL)
    fprintf(stderr, "%s: ssh failing\n", nodeName);
  else if ((bad = Match(alerts, "missing slice", FALSE)) != NULL &&
	   (bad = Match(bad, sliceName, TRUE)) != NULL)
    fprintf(stderr, "%s: missing slice %s\n", nodeName, sliceName);
  else if ((bad = Match(alerts, "extra slice", FALSE)) != NULL &&
	   (bad = Match(bad, sliceName, TRUE)) != NULL)
    fprintf(stderr, "%s: extra slice %s\n", nodeName, sliceName);
  else if ((bad = Match(alerts, "failed ssh.806", FALSE)) != NULL)
    fprintf(stderr, "%s: administrative ssh failing\n", nodeName);
  else {
    result = SUCCESS;
    if ((bad = Match(alerts, "failed http test", FALSE)) != NULL)
      fprintf(stderr, "%s: node http failing - not fatal, but suspicious\n", 
	      nodeName);
  }
  free(bad);
  free(alerts);
  return(result);
}
#endif
/*-----------------------------------------------------------------*/
static void
TestNode(char *nodeName, char *sliceName)
{
  char uri[256];
  char *buf;
  int retCode;
  char *temp;
  int result = FAILURE;

  if (strchr(nodeName, '@') != NULL) {
    fprintf(stderr, "%s: node has @ sign - skipping\n", nodeName);
    return;
  }
  
  if (strchr(nodeName, '.') == NULL) {
    fprintf(stderr, "%s: node name missing dots - skipping\n", nodeName);
    return;
  }

  if (gethostbyname(nodeName) == NULL) {
    fprintf(stderr, "%s: name lookup failed\n", nodeName);
    return;
  }

  if (CoMonTests(nodeName) == SUCCESS)
    result = SUCCESS;

#ifdef PSEPR_ALIVE
  if (AlertTests(nodeName, sliceName) == SUCCESS)
    result = SUCCESS;
#endif

  if (result != SUCCESS)
    return;

#ifndef PSEPR_ALIVE
  return;
#endif

  sprintf(uri, "php/feed/node_detail.php?node=%s", nodeName);
  buf = GetURL("www.psepr.org", 80, uri, 30, &retCode);
  if (buf == NULL || retCode < 1) {
    fprintf(stderr, "%s: failed getting node info\n", nodeName);
    free(buf);
    return;
  }

  if ((temp = strstr(buf, "idsl_monitor")) == NULL ||
      (temp = strstr(temp, "slices")) == NULL) {
    fprintf(stderr, "%s: ERROR - couldn't find list of slices\n", nodeName);
    free(buf);
    return;
  }
  else if ((temp = strstr(temp, sliceName)) == NULL) {
    fprintf(stderr, "%s: could not find created slice %s\n", 
	    nodeName, sliceName);
    free(buf);
    return;
  }

  fprintf(stderr, "%s: slice %s exists\n", nodeName, sliceName);
  free(buf);
  return;
}
/*-----------------------------------------------------------------*/
int
main(int argc, char *argv[])
{
  char *sliceName;
  int i;
  int sliceStart = 2;
  char *home;

  fprintf(stderr, 
	  "cotest version %s - "
	  "see http://codeen.cs.princeton.edu/cotest for updates\n", 
	  VERSION);

  if (argc < 2) {
    if ((sliceName = getenv("MQ_SLICE")) == NULL) {
      fprintf(stderr, "%s slice_name fully.qualified.domain.name ...\n", 
	      argv[0]);
      exit(-1);
    }
  }
  else
    sliceName = argv[1];

  if (sliceName == NULL || strchr(sliceName, '.') != NULL) {
    if ((sliceName = getenv("MQ_SLICE")) == NULL) {
      fprintf(stderr, "ERROR: your slice name contains a dot\n");
      exit(-1);
    }
    fprintf(stderr, "no slice specified on command line - using %s\n", 
	    sliceName);
    sliceStart--;
  }

  if (strchr(sliceName, '@') != NULL) {
    fprintf(stderr, "ERROR: your slice name contains @\n");
    exit(-1);
  }

  if (strchr(sliceName, '_') == NULL)
    fprintf(stderr, "WARNING: slice names generally contain underscores\n");

  if ((home = getenv("HOME")) == NULL) {
    fprintf(stderr, "ERROR: can't get environment variable HOME\n");
    exit(-1);
  }

  sprintf(cacheDir, "%s/cotest_cache", home);
  if ((mkdir(cacheDir, S_IRWXU) != 0) && errno != EEXIST) {
    perror("making cache dir");
    exit(-1);
  }
  if (access(cacheDir, R_OK | W_OK | X_OK) != 0) {
    fprintf(stderr, "can't access %s\n", cacheDir);
    exit(-1);
  }

#ifdef PSEPR_ALIVE
  allAlerts = FetchCacheableFile(cacheDir, "alert.csv", 3600, 
				 "www.psepr.org", "php/feed/alert.csv");
#endif
  if ((motd = FetchCacheableFile(cacheDir, "motd.txt", 3600, 
				 "codeen.cs.princeton.edu", 
				 "cotest/motd.txt")) != NULL) {
    motd = strstr(motd, "Status:");
  }
  if (motd != NULL)
    fprintf(stderr, "%s\n", motd);

  if ((servName = FetchCacheableFile(cacheDir, "server.txt", 3600, 
				 "codeen.cs.princeton.edu", 
				 "cotest/server.txt")) != NULL) {
    if ((servName = strstr(servName, "CoMonServer: ")) != NULL) {
      char *temp;
      servName += strlen("CoMonServer: ");
      for (temp = servName; *temp; temp++) {
	if (isspace(*temp)) {
	  *temp = '\0';
	  break;
	}
      }
      fprintf(stderr, "Using server: %s\n", servName);
    }
  }
  if (servName == NULL) {
    fprintf(stderr, "could not determine CoMon server\n");
    exit(-1);
  }

  sliceName = StrdupLower(sliceName);

  if (argc > sliceStart) {
    for (i = sliceStart; i < argc; i++) {
      char *nodeName = argv[i];
      
      TestNode(StrdupLower(nodeName), sliceName);
    }
  }
  else {
    char *nodeName;
    while ((nodeName = GetLowerNextLine(stdin)) != NULL)
      TestNode(nodeName, sliceName);
  }

  return(0);
}
/*-----------------------------------------------------------------*/
